# copyright 2003-2014 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
"""this module contains base classes and utilities for integration with running
http server
"""

import http.client
import random
import socket
from urllib.parse import urlparse

from webtest.http import StopableWSGIServer

from cubicweb.pyramid.test import _BasePyramidCWTest


def get_available_port(ports_scan):
    """return the first available port from the given ports range

    Try to connect port by looking for refused connection (111) or transport
    endpoint already connected (106) errors

    Raise a RuntimeError if no port can be found

    :type ports_range: list
    :param ports_range: range of ports to test
    :rtype: int

    .. see:: :func:`test.test_support.bind_port`
    """
    ports_scan = list(ports_scan)
    random.shuffle(ports_scan)  # lower the chance of race condition
    for port in ports_scan:
        try:
            s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            s.connect(("localhost", port))
        except socket.error as err:
            if err.args[0] in (111, 106):
                return port
        finally:
            s.close()
    raise RuntimeError(
        "get_available_port([ports_range]) cannot find an available port"
    )


class CubicWebServerTC(_BasePyramidCWTest):
    """Base class for running a test web server."""

    ports_range = range(7000, 8000)

    def start_server(self):
        app = self._generate_pyramid_config().make_wsgi_app()

        self.httpd = StopableWSGIServer.create(app, port=self.config["port"])
        self.httpd.wait()

        parseurl = urlparse(self.config["base-url"])
        assert parseurl.port == self.config["port"], (
            self.config["base-url"],
            self.config["port"],
        )

        self._web_test_cnx = http.client.HTTPConnection(
            parseurl.hostname, parseurl.port
        )
        self._ident_cookie = None

    def stop_server(self, timeout=15):
        if self._web_test_cnx is None:
            self.web_logout()
            self._web_test_cnx.close()
        self.httpd.shutdown()

    def web_login(self, user=None, passwd=None):
        """Log the current http session for the provided credential

        If no user is provided, admin connection are used.
        """
        if user is None:
            user = self.admlogin
            passwd = self.admpassword
        if passwd is None:
            passwd = user
        response = self.web_get("login")
        cookie = {
            x.split("=")[0]: x.split("=")[1]
            for x in "; ".join(response.headers.get_all("Set-Cookie")).split("; ")
            if "=" in x
        }
        self._ident_cookie = response.getheader("Set-Cookie")
        assert self._ident_cookie, self._ident_cookie
        response = self.web_request(
            "login",
            body=f"?__login={user}&__password={passwd}&csrf_token={cookie['csrf_token']}",
            method="POST",
        )
        assert response.status == http.client.OK, response.status
        return True

    def web_logout(self, user="admin", pwd=None):
        """Log out current http user"""
        if self._ident_cookie is not None:
            self.web_get("logout")
        self._ident_cookie = None

    def web_request(self, path="", method="GET", body=None, headers=None):
        """Return an http.client.HTTPResponse object for the specified path

        Use available credential if available.
        """
        if headers is None:
            headers = {}
        if self._ident_cookie is not None:
            assert "Cookie" not in headers
            headers["Cookie"] = self._ident_cookie
        self._web_test_cnx.request(method, "/" + path, headers=headers, body=body)
        response = self._web_test_cnx.getresponse()
        response.body = response.read()  # to chain request
        response.read = lambda: response.body
        return response

    def web_get(self, path="", body=None, headers=None):
        return self.web_request(path=path, body=body, headers=headers)

    def setUp(self):
        super().setUp()
        port = get_available_port(self.ports_range)
        self.config.global_set_option("port", port)  # force rewrite here
        self.config.global_set_option("base-url", "http://127.0.0.1:%d/" % port)
        # call load_configuration again to let the config reset its datadir_url
        self.config.load_configuration()
        self.start_server()

    def tearDown(self):
        self.stop_server()
        super().tearDown()
